/**
*
* @authors Andrei Novikov (pyclustering@yandex.ru)
* @date 2014-2019
* @copyright GNU Public License
*
* GNU_PUBLIC_LICENSE
*   pyclustering is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   pyclustering is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
*/

#include <pyclustering/differential/differ_factor.hpp>


namespace pyclustering {

namespace differential {


const double factor::A2 = 1.0 / 4.0;
const double factor::B2 = 1.0 / 4.0;

const double factor::A3 = 3.0 / 8.0;
const double factor::B3 = 3.0 / 32.0;
const double factor::C3 = 9.0 / 32.0;

const double factor::A4 = 12.0 / 13.0;
const double factor::B4 = 1932.0 / 2197.0;
const double factor::C4 = -7200.0 / 2197.0;
const double factor::D4 = 7296.0 / 2197.0;

const double factor::A5 = 1.0;
const double factor::B5 = 439.0 / 216.0;
const double factor::C5 = -8.0;
const double factor::D5 = 3680.0 / 513.0;
const double factor::E5 = -845.0 / 4104.0;

const double factor::A6 = 1.0 / 2.0;
const double factor::B6 = -8.0 / 27.0;
const double factor::C6 = 2.0;
const double factor::D6 = -3544.0 / 2565.0;
const double factor::E6 = 1859.0 / 4104.0;
const double factor::F6 = -11.0 / 40.0;

const double factor::N1 = 25.0 / 216.0;
const double factor::N3 = 1408.0 / 2565.0;
const double factor::N4 = 2197.0 / 4104.0;
const double factor::N5 = -1.0 / 5.0;

const double factor::R1 = 1.0 / 360.0;
const double factor::R3 = -128.0 / 4275.0;
const double factor::R4 = -2197.0 / 75240.0;
const double factor::R5 = 1.0 / 50.0;
const double factor::R6 = 2.0 / 55.0;


}

}