/**
*
* @authors Andrei Novikov (pyclustering@yandex.ru)
* @date 2014-2019
* @copyright GNU Public License
*
* GNU_PUBLIC_LICENSE
*   pyclustering is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   pyclustering is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
*/


#pragma once


namespace pyclustering {

namespace differential {


class factor {
public:
    const static double A2;
    const static double B2;

    const static double A3;
    const static double B3;
    const static double C3;

    const static double A4;
    const static double B4;
    const static double C4;
    const static double D4;

    const static double A5;
    const static double B5;
    const static double C5;
    const static double D5;
    const static double E5;

    const static double A6;
    const static double B6;
    const static double C6;
    const static double D6;
    const static double E6;
    const static double F6;

    const static double N1;
    const static double N3;
    const static double N4;
    const static double N5;

    const static double R1;
    const static double R3;
    const static double R4;
    const static double R5;
    const static double R6;
};


}

}